/* common.c */

/* cf random(3) */
#define _BSD_SOURCE

#include "df3util.h"

/* cf strcasecmp(3) */
#include <strings.h>


/* mkSourceFname(), mkTargetFname()
 * -------------------------------------------------------------------------- *
 * sources must always exist in CWD or {app_workdir}, and have correct perms.
 * targets must never exist in {app_workdir}.  can use absolute names.
 * the chosen extension is appended where required.
 * both funcs return false and the reason in {app_lastec} when they fail,
 * true and an allocated filname in {*fname} on success.
 * caller responsible for free()ing fnames.
 * TODO use errno when strdup() etc fail?
 */

bool mkSourceFname(char const * const arg,
                   bool const xpng,
                   bool const wr,
                   char **fname) {
  assert(!*fname);
  char buf[PATH_MAX] = {CHRNUL};
  char const * const extn = xpng ? ".png" : DF3_EXTENSION;
  size_t const n = strlen(arg);

  if (!n  ||  PTHNET < n) {
    app_lastec = ec_badFilename;
    return false;
  }
  strcpy(buf, arg);

  char *p = strrchr(arg, CHRDOT);
  if (!p  ||  strcasecmp(extn, p))
    strcat(buf, extn);

  if (!statFileExists(buf)) {
// TODO check snprintf return.
    char buf2[PATH_MAX] = {CHRNUL};
    snprintf(buf2, PATH_MAX, "%s%c%s", app_workdir ? app_workdir : ".",
            PTHSEP, buf);
    if (!statFileExists(buf2)) {
      app_lastec = ec_noFile;
      return false;
    }
    strcpy(buf, buf2);
  }

  bool wflag = false;
  if (!statFilePerms(buf, &wflag)) {
    app_lastec = ec_permsFile;
    return false;
  }
  else if (wr  &&  !wflag) {
    app_lastec = ec_permsFile;
    return false;
  }

  p = strdup(buf);
  if (!p) {
    app_lastec = ec_sysError;
    return false;
  }

  *fname = p;
  return true;
}


bool mkTargetFname(char const * const arg, bool const xpng, char **fname) {
  assert(!*fname);
  char buf[PATH_MAX] = {CHRNUL};
  char const * const extn = xpng ? ".png" : DF3_EXTENSION;
  size_t const n = strlen(arg);

  if (!n  ||  PTHNET < n) {
    app_lastec = ec_badFilename;
    return false;
  }

// TODO check snprintf return.
  if (PTHSEP == arg[0]  ||  CHRDOT == arg[0])
    strcpy(buf, arg);
  else
    snprintf(buf, PATH_MAX, "%s%c%s", app_workdir ? app_workdir : ".",
            PTHSEP, arg);

  char *p = strrchr(arg, CHRDOT);
  if (!p  ||  strcasecmp(extn, p))
    strcat(buf, extn);

  if (statFileExists(buf)) {
    app_lastec = ec_existFile;
    return false;
  }

  p = strdup(buf);
  if (!p) {
    app_lastec = ec_sysError;
    return false;
  }

  *fname = p;
  return true;
}


/* mkBufferXY()
 * -------------------------------------------------------------------------- *
 * selects the image column and row dimensions, according to {axis}.
 */

void mkBufferXY(df3Axes const axes,
                    char const axis,
                    size_t *cols,
                    size_t *rows) {
  switch (axis) {
    case 'x':  *cols = axes.z;  *rows = axes.y;  break;
    case 'y':  *cols = axes.x;  *rows = axes.z;  break;
    case 'z':  *cols = axes.x;  *rows = axes.y;  break;
    default:   APPPANIC;
  }
  return;
}


/* axisMax()
 * -------------------------------------------------------------------------- *
 */

size_t axisMax(df3Axes const axes, char const axis) {
  size_t n = 0;
  switch (axis) {
    case 'x':  n = axes.x;  break;
    case 'y':  n = axes.y;  break;
    case 'z':  n = axes.z;  break;
    default:   APPPANIC;
  }
  return n;
}


/* badAcol()
 * -------------------------------------------------------------------------- *
 * nb true when too large.
 */

bool badAcol(df3Axes const axes, char const axis, size_t const n) {
  return axisMax(axes, axis) <= n;
}


/* atOffset()
 * -------------------------------------------------------------------------- *
 * simple wrapper for addition/subtraction, args already known good.
 */

df3Coord atOffset(bool const plus, df3Coord const at, Offset const o) {
  df3Coord const x = plus ? (df3Coord){.x = (size_t)(o.x + (int)at.x),
                                       .y = (size_t)(o.y + (int)at.y),
                                       .z = (size_t)(o.z + (int)at.z)}
                          : (df3Coord){.x = (size_t)(o.x - (int)at.x),
                                       .y = (size_t)(o.y - (int)at.y),
                                       .z = (size_t)(o.z - (int)at.z)};
  return x;
}


/* isDefaultArg()
 * -------------------------------------------------------------------------- *
 * TODO inline?
 */

bool isDefaultArg(char const * const arg) {
  return 1 == strlen(arg)  &&  CHRDOT == arg[0];
}


/* getInpSrc()
 * -------------------------------------------------------------------------- *
 * expect three args "- axes bpv" input source spec (also accept dot),
 * cli format only if the -input option was given.
 * returns:  1 - valid input source spec, have input, axes + bpv set. 
 *          -1 - missing arg(s) or fail conversion/limit, or no -input,
 *               axes + bpv not set, error.
 *           0 - not input source spec.
 */

int getInpSrc( char * av[], df3Axes *axes, size_t *bpv) {
  bool const hyphen = 1 == strlen(av[0])  &&  CHRHYP == av[0][0];

  if (!(hyphen  ||  isDefaultArg(av[0])))
    return 0;
  else if (!(av[1]  &&  av[2])) {
    EMITEMSG(false, "missing argument(s)");
    app_lastec = ec_badUsage;
    return -1;
  }
  else if (itNone == app_inptype) {
    EMITEMSG(false, "missing -input option");
    app_lastec = ec_badUsage;
    return -1;
  }

  char const * const fmt = "%s axes value(s)";
  df3Axes a;
  if (!getArgCoord(av[1], &a)) {
    EMITEMSG(false, fmt, "bad");
    return -1;
  }
  else if (!df3_check_axes(a)) {
    EMITEMSG(false, fmt, "zero");
    return -1;
  }

  char const * const msg = "bad bpv value";
  uint32_t b = 0;
  if (!getArgValue(av[2], &b)) {
    EMITEMSG(false, "%s", msg);
    return -1;
  }
  else if (!df3_check_bpv((size_t)b)) {
    EMITEMSG(false, "%s", msg);
    return -1;
  }

  *axes = a;
  *bpv = (size_t)b;
  return 1;
}


/* isValidFillval()
 * -------------------------------------------------------------------------- *
 * true if {app_fillval} fits (or allowed to make fit) given bpv.
 */

bool isValidFillval(size_t const b) {
  if (-1L == app_fillval)
    return true;

  bool f = false;
  switch (b) {
    case 4:  f = UINT32_MAX >= app_fillval;  break;
    case 2:  f = UINT16_MAX >= app_fillval;  break;
    case 1:  f = UINT8_MAX >= app_fillval;   break;
    default: APPPANIC;
  }

  if (!f) {
    app_lastec = ec_badArgument;
    if (!app_warns) {
      EMITEMSG(false, "fill value too large");
      return f;
    }
    EMITEMSG(true, "fill value will be truncated");
    f = true;
  }

  return f;
}


/* isValidVoxel()
 * -------------------------------------------------------------------------- *
 */

bool isValidVoxel(size_t const bpv, uint32_t const val) {
  bool f = true;
  if (4 != bpv)
    f = val <= ((2 == bpv) ? UINT16_MAX : UINT8_MAX);
  return f;
}


/* inpVoxelValid()
 * -------------------------------------------------------------------------- *
 * for -input callbacks to check voxel is ok for given axes/bpv.
 */

bool inpVoxelValid(df3Axes const a,
                   size_t const b,
                   df3Coord const c,
                   uint32_t const v) {
  if (!containedIn(a, ZEROOFFS, c)) {
    app_lastec = ec_badOffset;
    EMITEMSG(false, "bad offset <%zu,%zu,%zu>", c.x, c.y, c.z);
    return false;
  }
  if (!isValidVoxel(b, v)) {
    app_lastec = ec_badVoxel;
    EMITEMSG(false, "bad voxel (%u) at <%zu,%zu,%zu>", v, c.x, c.y, c.z);
    return false;
  }
  return true;
}


/* getFillValue()
 * -------------------------------------------------------------------------- *
 * {app_fillval} known to be no larger than UINT32_MAX.
 */

uint32_t getFillValue(size_t const b) {
  uint32_t v = 0;
  switch (b) {
    case 4:
      if (-1L == app_fillval)
        v = (uint32_t)random();
      else
        v = (uint32_t)app_fillval;
      break;
    case 2:
      if (-1L == app_fillval)
        v = (uint32_t)random() & UINT16_MAX;
      else
        v = (uint32_t)app_fillval & UINT16_MAX;
      break;
    case 1:
      if (-1L == app_fillval)
        v = (uint32_t)random() & UINT8_MAX;
      else
        v = (uint32_t)app_fillval & UINT8_MAX;
      break;
    default:
      APPPANIC;
  }
  return v;
}


/* prepFilled()
 * -------------------------------------------------------------------------- *
 * filling new volumes before use simplifies things.
 */

bool prepFilled(char const * const tu, df3Handle *df3) {
  if (app_diagnose)
    EMITDMSG(tu, "preparing volume, %s value fill",
            0 > app_fillval ? "random" : "fixed");
  for (size_t z = 0; df3->axes.z != z; ++z)
    for (size_t y = 0; df3->axes.y != y; ++y)
      for (size_t x = 0; df3->axes.x != x; ++x) {
        df3Error ec = df3_put_voxel(df3, (df3Coord){x, y, z},
                getFillValue(df3->bpv));
        if (ecOk != ec) {
          EMITEMSG(false, "preparing volume (at <%zu,%zu,%zu>): %s",
                  x, y, z, df3_error_message(ec));
          app_lastec = le2ae(ec);
          return false;
        }
      }

  return true;
}


/* trimmedFn()
 * -------------------------------------------------------------------------- *
 * buffer with trimmed and prefixed, when necessary, filename.
 * NULL returns previous.
 */

#define TFNLEN 47

static char trimmedfnbuf[TFNLEN+4];

char const *trimmedFn(char const * const fn) {
  if (fn) {
    assert(strlen(fn));
    char const * const p = fn;
    size_t n = strlen(p);
    if (TFNLEN < n)
      n = n - TFNLEN;
    else
      n = 0;
    sprintf(trimmedfnbuf, "%s%s", n ? "..." : "", p + n);
  }

  return &trimmedfnbuf[0];
}

/* ------------------------------------------------------------------ *
 *  this file is part of 'df3util', part of the 'df3 tools' project.  *
 *  the content above is covered by the GPLv3+, see file COPYING.     *
 *  copyright (c) 2017 jr <creature.eternal@gmail.com>.               *
 *  all rights reserved.                                              *
 * ------------------------------------------------------------------ */
