/* vinters.c */

#include "df3util.h"

/* from axes and coords */
#define TOOFFS(a_) (Offset){(int)((a_).x), (int)((a_).y), (int)((a_).z)}

static Offset const lolim = { -65533, -65533, -65533 },
                    hilim = { UINT16_MAX, UINT16_MAX, UINT16_MAX };

static inline bool isLess(Offset const o1, Offset const o2) {
  return (o1.x < o2.x  &&  o1.y < o2.y  &&  o1.z < o2.z);
}

static inline bool isNeg(Offset const o) {
  return (0 > o.x  ||  0 > o.y  ||  0 > o.z);
}

static int cmpOffs(Offset const o1, Offset const o2) {
  if (o1.x < o2.x  ||  o1.y < o2.y  ||  o1.z < o2.z)
    return -1;
  else if (o1.x == o2.x  &&  o1.y == o2.y  &&  o1.z == o2.z)
    return 0;
  else
    return 1;
}

static inline bool legalOffs(Offset const o) {
  if (-1 == cmpOffs(o, lolim))
    return false;
  else if (-1 != cmpOffs(o, hilim))
    return false;
  else
    return true;
}

static Offset addOffs(Offset const o1, Offset const o2, bool *f) {
  Offset const res = (Offset){o1.x + o2.x, o1.y + o2.y, o1.z + o2.z};
  *f = legalOffs(res);
  return res;
}


/* containedIn()
 * -------------------------------------------------------------------------- *
 * true if offset+coord in volume.
 */

bool containedIn(df3Axes const a, Offset const o, df3Coord const c) {
  bool rf = false;
  Offset const t = addOffs(o, TOOFFS(c), &rf);
  if (!rf)
    return false;
  else if (isNeg(t))
    return false;
  else if (isLess(t, TOOFFS(a)))
    return true;
  else
    return false;
}


/* volsOverlap()
 * -------------------------------------------------------------------------- *
 * true if offset+subv fully contained in vol, or at least part (with -warn).
 */

bool volsOverlap(df3Axes const vol,
                 Offset const o,
                 df3Coord const subv,
                 bool *inside) {
  df3Coord const t = {subv.x - 1, subv.y - 1, subv.z - 1};
  *inside = false;

  bool rf = false;
  Offset junk AUU = addOffs(o, TOOFFS(t), &rf);
  if (!rf) {
    app_lastec = ec_badOffset;
    return false;
  }

  int rc = 0;
  rc += containedIn(vol, o, (df3Coord){  0,   0,   0}) ? 1 : 0;
  rc += containedIn(vol, o, (df3Coord){t.x,   0,   0}) ? 1 : 0;
  rc += containedIn(vol, o, (df3Coord){  0, t.y,   0}) ? 1 : 0;
  rc += containedIn(vol, o, (df3Coord){t.x, t.y,   0}) ? 1 : 0;
  rc += containedIn(vol, o, (df3Coord){  0,   0, t.z}) ? 1 : 0;
  rc += containedIn(vol, o, (df3Coord){t.x,   0, t.z}) ? 1 : 0;
  rc += containedIn(vol, o, (df3Coord){  0, t.y, t.z}) ? 1 : 0;
  rc += containedIn(vol, o, (df3Coord){t.x, t.y, t.z}) ? 1 : 0;

  if (8 == rc)  /* fully inside */
    *inside = true;
  else if (!rc) {  /* disjoint */
    app_lastec = ec_badOffset;
    return false;
  }
  else if (!app_warns) {
    app_lastec = ec_badOffset;
    return false;
  }

  return true;
}


/* atCLEinters()
 * -------------------------------------------------------------------------- *
 * copy less equal: true if at in intersection, false otherwise.
 * if true, at set to ref subv.
 * only calculate the intersection etc if vol/o/subv different from previous
 * call, 'prev_*' keep "state".
 * NB code relies on caller having checked volumes overlapping.
 */

static df3Axes  prev_vol;  /* args last call */
static Offset   prev_o;
static df3Coord prev_subv;

static df3Coord prev_ivi;  /* volume intersection */
static Offset prev_ipt1,   /* begin + end of intersection */
              prev_ipt2;   /* frame of ref: vol */
static Offset prev_add;    /* addend changes frame of ref */

static inline void pt1axis(int const o, int *adj, size_t *ivi, int *pt) {
  if (0 > o) {
    *adj = abs(o);
    *ivi -= (size_t)*adj;
  }
  else if (0 < o) {
    *adj = -o;
    *pt = o;
  }
}

static inline void pt2axis(size_t const a, size_t *ivi, int *pt) {
  while ((int const)a <= *pt) {
    *pt -= 1;
    *ivi -= 1;
  }
}

static bool calcIntersect(df3Axes const vol,
                          Offset const o,
                          df3Coord const subv) {
  bool inside = false;
  if (!volsOverlap(vol, o, subv, &inside))
    return false;

  prev_vol = vol;
  prev_o = o;
  prev_subv = subv;

  prev_ivi = subv;
  prev_ipt1 = prev_add = ZEROOFFS;
  bool rf = false;
  prev_ipt2 = addOffs(o, TOOFFS(subv), &rf);
  if (!rf) APPPANIC;
  prev_ipt2.x -= 1;  prev_ipt2.y -= 1;  prev_ipt2.z -= 1;

  pt1axis(o.x, &prev_add.x, &prev_ivi.x, &prev_ipt1.x);
  pt1axis(o.y, &prev_add.y, &prev_ivi.y, &prev_ipt1.y);
  pt1axis(o.z, &prev_add.z, &prev_ivi.z, &prev_ipt1.z);
  assert(!isNeg(prev_ipt1));

  pt2axis(vol.x, &prev_ivi.x, &prev_ipt2.x);
  pt2axis(vol.y, &prev_ivi.y, &prev_ipt2.y);
  pt2axis(vol.z, &prev_ivi.z, &prev_ipt2.z);
  assert(!isNeg(prev_ipt2));
  assert(df3_check_axes(prev_ivi));

  if (app_diagnose) {
    int const w = 14;
    EMITDMSG("calc","intersect\n"
                    "%*s: %zu,%zu,%zu\n"
                    "%*s: %zu,%zu,%zu\n"
                    "%*s: %d,%d,%d\n"
                    "%*s: %s\n"
                    "%*s: %d,%d,%d\n"
                    "%*s: %d,%d,%d\n"
                    "%*s: %zu,%zu,%zu\n"
                    "%*s: %d,%d,%d",
            w, "1st vol", prev_vol.x, prev_vol.y, prev_vol.z,
            w, "2nd vol", prev_subv.x, prev_subv.y, prev_subv.z,
            w, "offset", prev_o.x, prev_o.y, prev_o.z,
            w, "overlap", inside ? "full" : "part",
            w, "i pt1", prev_ipt1.x, prev_ipt1.y, prev_ipt1.z,
            w, "i pt2", prev_ipt2.x, prev_ipt2.y, prev_ipt2.z,
            w, "i vol", prev_ivi.x, prev_ivi.y, prev_ivi.z,
            w, "addend", prev_add.x, prev_add.y, prev_add.z);
  }

  return true;
}

static inline bool argsDiffer(df3Axes const vol,
                              Offset const o,
                              df3Coord const subv) {
  return cmpOffs(o, prev_o)  ||  cmpOffs(TOOFFS(vol),TOOFFS(prev_vol))
         ||  cmpOffs(TOOFFS(subv),TOOFFS(prev_subv));
}

bool atCLEinters(df3Axes const vol,
                df3Coord *at,
                Offset const o,
                df3Coord const subv) {

  if (argsDiffer(vol, o, subv))
    if (!calcIntersect(vol, o, subv))
      return false;

  if (-1 == cmpOffs(TOOFFS(*at), prev_ipt1))
    return false;
  else if (1 == cmpOffs(TOOFFS(*at), prev_ipt2))
    return false;
  
  bool rf = false;
  Offset const xat = addOffs(TOOFFS(*at), prev_add, &rf);
  if (!rf) APPPANIC;
  assert(!isNeg(xat));
  
  df3Coord const x = (df3Coord){(size_t)xat.x, (size_t)xat.y, (size_t)xat.z};
  if (!containedIn(subv, ZEROOFFS, x))  /* final sanity check */
    return false;
  *at = x;

  return true;
}


/* atCGTinters()
 * -------------------------------------------------------------------------- *
 * copy greater than: as merge but use src rather than axes.
 */

bool atCGTinters(df3Axes const src,
                df3Coord *at,
                Offset const offs,
                df3Coord const tgt) {
  return atMinters(src, at, offs, tgt);
}


/* atMinters()
 * -------------------------------------------------------------------------- *
 * merge: true if at in intersection, false otherwise;  if true, at changed
 * to ref tgt.  calc intersect tgt with (offs + src), only if src/offs/tgt
 * different from previous call, 'prev_*' keep "state".
 * NB code relies on caller having checked volumes overlapping.
 */

static df3Axes  prev_src;  /* args last call */
static Offset   prev_offs;
static df3Coord prev_tgt;


static inline bool argsMdiffer(df3Axes const src,
                              Offset const offs,
                              df3Coord const tgt) {
  return cmpOffs(offs, prev_offs)
         ||  cmpOffs(TOOFFS(src),TOOFFS(prev_src))
         ||  cmpOffs(TOOFFS(tgt),TOOFFS(prev_tgt));
}

bool atMinters(df3Axes const src,
                df3Coord *at,
                Offset const offs,
                df3Coord const tgt) {

  assert(containedIn(src, ZEROOFFS, *at));

  /* NB reversed src/tgt */
  if (argsMdiffer(src, offs, tgt)) {
    if (!calcIntersect(tgt, offs, src))
      return false;
    prev_src = src;
    prev_offs = offs;
    prev_tgt = tgt;
    prev_add = (Offset){-prev_add.x, -prev_add.y, -prev_add.z};
  }

  bool rf = false;
  Offset const xat = addOffs(TOOFFS(*at), prev_add, &rf);
  if (!rf) APPPANIC;

  if (-1 == cmpOffs(xat, prev_ipt1)  ||  1 == cmpOffs(xat, prev_ipt2))
    return false;

  assert(!isNeg(xat));  // redundant?

  *at = (df3Coord){(size_t)xat.x, (size_t)xat.y, (size_t)xat.z};

  return true;
}

/* ------------------------------------------------------------------ *
 *  this file is part of 'df3util', part of the 'df3 tools' project.  *
 *  the content above is covered by the GPLv3+, see file COPYING.     *
 *  copyright (c) 2017 jr <creature.eternal@gmail.com>.               *
 *  all rights reserved.                                              *
 * ------------------------------------------------------------------ */
